# -*- coding: utf-8 -*-
# Wendelin.bigfile | WCFS part of BigFile ZODB backend
# Copyright (C) 2014-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# cython: language_level=2
# distutils: language=c++

"""XXX"""

cdef extern from "wcfs/internal/wcfs.h":
    pass

#from wendelin.bigfile import BigFile
# XXX C++ does not understand -fplan9-extensions
"""
cdef extern from "bigfile/_bigfile.h":
    ctypedef extern class wendelin.bigfile._bigfile.BigFile[object PyBigFile]:
        pass
"""
# FIXME hack, keep in sync with PyBigFile
cdef extern from *:
    """
    #include <wendelin/bigfile/file.h>
    #include <wendelin/bigfile/virtmem.h>

    struct cxxPyBigFile {
        PyObject pyobj;
        BigFile  file;
    };
    """
    struct cxxPyBigFile:
        pass
    ctypedef extern class wendelin.bigfile._bigfile.BigFile[object cxxPyBigFile]:
        pass

from wcfs.internal cimport _wcfs as wcfs
from golang cimport nil

from ZODB.utils import u64


# _ZBigFile is helper for ZBigFile - just redirect loadblk/storeblk back
# (because it is not possible to inherit from both Persistent and BigFile at
#  the same time - see below)
#
# XXX + wcfs
cdef class _ZBigFile(BigFile):
    cdef object     zself   # reference to ZBigFile
    cdef wcfs.Oid   foid    # = .zself._p_oid
    cdef wcfs.Conn  wconn   # WCFS connection   XXX -> zconn instead ?  XXX wconn is shared between files

    # XXX Cython does not allow __new__ nor to change arguments passed to __cinit__ / __init__
    @staticmethod
    def _new(zself, blksize):
        cdef _ZBigFile obj = _ZBigFile.__new__(_ZBigFile, blksize)
        obj.zself = zself
        obj.foid  = -1  # Note: zself._p_oid could be yet None for newly created ZBigFile
        obj.wconn = nil # XXX
        return obj

    def __dealloc__(_ZBigFile zf):
        zf.wconn = nil


    # redirect load/store/mapper to main class
    def loadblk(self, blk, buf):    return self.zself.loadblk(blk, buf)
    def storeblk(self, blk, buf):   return self.zself.storeblk(blk, buf)
    #def blkmmapper(self):           return self.zself.blkmmapper()

    # blkmapper serves PyBigFile and gives it XXX ...
    #
    # it is called from PyBigFile.fileh_open(mmap_overlay=True) and .zfile
    # should be already associaed with jar and have oid.
    def blkmapper(_ZBigFile zf):
        if zf.wconn == nil:
            zconn = zf.zself._p_jar
            assert zconn is not None
            # XXX locking? or rely on that ZODB objects for must be used from under 1 thread only?
            zf.foid = u64(zf.zself._p_oid)

            # XXX join zconn to wconn
            pywconn = pywconnOf(zconn)
            zf.wconn = pywconn.XXX


# functions that we give to virtmem bigfile_ops .mmap*
cdef nogil:

    void* virt_mmap_setup_read(VMA *vma, BigFile *file, blk_t blk, size_t blklen) {
        FileH fileh; # XXX = ... somehow from file
        Mapping mmap;
        error err;

        tie(mmap, err) = fileh->mmap(blk, blklen, vma);
        if (err != nil)
            panic("TODO");  # XXX

        mmap->incref(); # vma->mmap_overlay_server is keeping ref to mmap
        vma->mmap_overlay_server = mmap._ptr();

        # XXX here? -> in fileh->mmap? don't set at all here?
        vma->addr_start = (uintptr_t)mmap->mem_start;
        vma->addr_stop  = (uintptr_t)mmap->mem_stop;

        return (void*)vma->addr_start;  // XXX kill if we set vma->addr_*
    }

    int virt_remmap_blk_read(VMA *vma, BigFile *file, blk_t blk) {
        _Mapping *mmap = static_cast<_Mapping*>(vma->mmap_overlay_server);
        // XXX use file?

        mmap->remmap_blk(blk);  // XXX err
        return 0;
    }


    void virt_munmap(VMA *vma, BigFile *file) {
        Mapping mmap = adoptref(static_cast<_Mapping*>(vma->mmap_overlay_server)); // NOTE taking 1 ref back
        vma->mmap_overlay_server = NULL;

        mmap->unmap();

        // XXX close mmap->fileh
    }




# pywconnOf establishes and returns (py) wcfs.Conn associated with zconn.
# XXX wconn is maintained and kept in sync with zconn.
# XXX simultaneous call?
# XXX move -> .py ?
cdef pywconnOf(zconn):
    assert isinstance(zconn, zodbConnection)
    assert zconn.opened     # XXX needed

    # XXX locking
    wconn = getattr(zconn, '_wcfs_wconn', None)
    if wconn is not None:
        return wconn

    # zconn is not yet associated with wconn
    zstor = zconn.db().storage
    zurl = wcfs.zstor_2zurl(zstor)
    wc = wcfs.join(zurl)
    wconn = wc.connect(zconn_at(zconn))

    zconn._wcfs_wconn = wconn
    return wconn
