# -*- coding: utf-8 -*-
# Copyright (C) 2018-2019  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

# cython: language_level=2
# distutils: language=c++

# XXX doc

from golang cimport chan, structZ, string, error, refptr
from golang cimport context

from libc.stdint cimport int64_t, uint64_t
from libcpp.utility cimport pair
from libcpp.vector cimport vector


# XXX -> pygolang
cdef extern from "wcfs/internal/wcfs_misc.h" namespace "io" nogil:
    error EOF "io::EOF_"
    error ErrUnexpectedEOF

cdef extern from "wcfs/internal/wcfs_misc.h" namespace "zodb" nogil:
    ctypedef uint64_t Tid
    ctypedef uint64_t Oid

cdef extern from "wcfs/internal/wcfs_misc.h" namespace "wcfs" nogil:
    const Tid TidHead


cdef extern from "wcfs/internal/wcfs_watchlink.h" namespace "wcfs" nogil:
    cppclass _WatchLink:
        error close()
        error closeWrite()
        pair[string, error] sendReq(context.Context ctx, const string &req)
        error recvReq(context.Context ctx, PinReq *prx)
        error replyReq(context.Context ctx, const PinReq *req, const string& reply);

        vector[string] fatalv
        chan[structZ]  rx_eof

    cppclass WatchLink (refptr[_WatchLink]):
        # WatchLink.X = WatchLink->X in C++
        error               close       "_ptr()->close"     ()
        error               closeWrite  "_ptr()->closeWrite"()
        pair[string, error] sendReq     "_ptr()->sendReq"   (context.Context ctx, const string &req)
        error               recvReq     "_ptr()->recvReq"   (context.Context ctx, PinReq *prx)
        error               replyReq    "_ptr()->replyReq"  (context.Context ctx, const PinReq *req, const string& reply);

        vector[string]      fatalv      "_ptr()->fatalv"
        chan[structZ]       rx_eof      "_ptr()->rx_eof"

    cppclass PinReq:
        Oid     foid
        int64_t blk
        Tid     at
        string  msg

    error _twlinkwrite(WatchLink wlink, const string& pkt)


cdef extern from "wcfs/internal/wcfs.h" namespace "wcfs" nogil:
    cppclass WCFS:
        string  mountpoint

        pair[WatchLink, error] _openwatch()
        pair[Conn, error]      connect(Tid at)

    cppclass _Conn:
        pair[FileH, error]  open(Oid foid)
        error               close()

    cppclass Conn (refptr[_Conn]):
        # Conn.X = Conn->X in C++
        pair[FileH, error]  open        "_ptr()->open"      (Oid foid)
        error               close       "_ptr()->close"     ()

    cppclass _FileH:
        # XXX add mmap?
        pass

    cppclass FileH (refptr[_FileH]):
        # FileH.X = FileH->X in C++
        # XXX add mmap?
        pass
